#!/usr/bin/env python
#
#	Copyright 2013-2014, Marten de Vries
#
#	Licensed under the Apache License, Version 2.0 (the "License");
#	you may not use this file except in compliance with the License.
#	You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
#	Unless required by applicable law or agreed to in writing, software
#	distributed under the License is distributed on an "AS IS" BASIS,
#	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#	See the License for the specific language governing permissions and
#	limitations under the License.

import subprocess
import shlex
import os
import unittest
import collections
import json
import re
import sys
import atexit
import pouchdb.tests.utils
import glob

def do(command, pipeOutput=False):
	if pipeOutput:
		stdout = subprocess.PIPE
		stderr = subprocess.PIPE
	else:
		stdout = None
		stderr = None
	return subprocess.Popen(shlex.split(command), stdin=subprocess.PIPE, stdout=stdout, stderr=stderr)

def editingPackageFiles():
	for path in packageFiles:
		with open(path) as f:
			data = json.load(f, object_pairs_hook=collections.OrderedDict)
		yield data
		jsonData = json.dumps(data, indent=4, separators=(",", ": ")).replace(" " * 4, "\t") + "\n"
		with open(path, "w") as f:
			f.write(jsonData)

#setup coverage JS
os.chdir("js/python-pouchdb-js")

packageFiles = glob.glob("../*/package.json")

#temporary add the transform to some package.json's - es3ify seems to
#conflict with coverify and since the first isn't necessary for the
#webkit environment, it's just removed temporarily.
for package in editingPackageFiles():
	package["browserify"]["transform"] = [
		["coverify", {"output": "coverage_info"}]
	]

def changePackageFilesBack():
	for package in editingPackageFiles():
		package["browserify"]["transform"] = ["es3ify"]
		if package["name"] == "python-pouchdb-js":
			package["browserify"] = {}


atexit.register(changePackageFilesBack)

if do("npm run build-js").wait() != 0:
	sys.exit()

os.rename("../../pouchdb/bundle-full.js", "../../pouchdb/bundle.js")

with open("../../pouchdb/bundle.js") as f:
	data = "var info = []; function coverage_info(d) {info.push(d); }; " + f.read()

with open("../../pouchdb/bundle.js", "w") as f:
	f.write(data)

#Run tests; saving the coverage's js result data.
data = []
def tearDownHandler(env):
	info = env.context.evalJs("return window.info || [];")
	data.extend(info)
pouchdb.tests.utils.tearDownHandlers.add(tearDownHandler)

program = unittest.main(module="pouchdb.tests", exit=False, failfast=True)
if not program.result.wasSuccessful():
	sys.exit()

#only unique rows - this is necessary to not only use the data from the
#last test run.
data = list(collections.OrderedDict.fromkeys(data))

#replace symlinks with their real equivalents - also to filter out
#duplicates
def toRealPath(match):
	return '"%s"' % os.path.realpath(match.group(1))
data = [re.sub(r'"(.*?)"', toRealPath, line) for line in data]

#again only unique rows - the symlink removal might have created new
#duplicates. Not sure if that does any harm, but better safe than sorry.
data = list(collections.OrderedDict.fromkeys(data))

#run the coverage tool
data = do("coverify --color false", pipeOutput=True).communicate("\n".join(data))[1]

def excluded(row):
	return any([
		"designdoc.js" in row,
		"coverage: ignore" in row,
	])

rows = data.split("#")
del rows[-1]
filteredRows = [row for row in rows if not excluded(row)]
filteredRows = [row for row in filteredRows if row]

print "\nCOVERAGE REPORT:"
print "#".join(filteredRows)
if len(filteredRows):
	print "COUNT:", len(filteredRows)
else:
	print "COVERAGE 100%"
